; Programa: Ejemplo de agente basico humano para jugar al sokoban
;           ver http://e-contento.com/sokoban/
;
; Asignatura: Introduccin a la Inteligencia Artificial
;             CSI - ULL
; Prctica: 5a (sistemas de produccin)
; Profesor: Patricio Garca
; Fecha: 27/5/2005
; Comentarios:
;              Hace uso del package sokoban
;              (load "sokoban")
;

(use-package "sokoban")


(defmacro robotfil (est)
  `(first (soko-robot ,est)))

(defmacro robotcol (est)
  `(second (soko-robot ,est)))

(defmacro en-rel-robot (est inci incj)
  `(aref (soko-mat ,est)
	(+ (robotfil ,est) ,inci)
	(+ (robotcol ,est) ,incj)))

(defvar mov2inc
  (make-array '(4 2) :initial-contents '((-1 0) (1 0) (0 1) (0 -1)))
  "Array de incrementos relativos para movimientos n,s,e y o")

(defun mov-val (mov est)
  "Chequea si es valido el movimiento mov en el estado est"
  (let ((inci 0) (incj 0))
    (if (not (member mov '(1 2 3 4)))
	(progn (print "Movimiento desconocido") nil)
      (progn
	; Calcula incrementos relativos a mover en filas y columnas
        (setf inci (aref mov2inc (1- mov) 0)
              incj (aref mov2inc (1- mov) 1))
	; Chequea si se puede mover (cumpla una de las condiciones)
	(or (not (en-rel-robot est inci incj))
	    ; Condicion 1: no hay nada en donde se quiere mover
	    (and (eq (en-rel-robot est inci incj)
		     *blq-soko*)
		 (not (en-rel-robot est (+ inci inci) (+ incj incj)))))))))
            ; Condicion 2: hay un bloque y nada despues del bloque,
            ;              se puede empujar el bloque



(defun mueve (mov est)
  "Realiza movimiento mov en el estado est"
  (let ((inci (aref mov2inc (1- mov) 0)) 
	(incj (aref mov2inc (1- mov) 1)))
    (cond
     ((not (en-rel-robot est inci incj))
      ; Movimiento sin empujar
      (setf (robotfil est) (+ (robotfil est) inci))
      (setf (robotcol est) (+ (robotcol est) incj))
      est)
     (t
      ; Movimiento empujando un bloque
      (setf (en-rel-robot est inci incj) nil)
      (setf (en-rel-robot est (+ inci inci) (+ incj incj)) *blq-soko*)
      (setf (robotfil est) (+ (robotfil est) inci))
      (setf (robotcol est) (+ (robotcol est) incj))
      est))))

(defun car2num (car)
  "Transforma el caracter de entrada al numero de movimiento"
  (case car
    ((n k) 1)
    ((j s) 2)
    ((l e) 3)
    ((h o) 4)))


; Agente humano
; -------------

(defun agente-humano ()
  "Agente humano para jugar al sokoban"
  (do ((est (gen-soko) (mueve mov est))
       (solucion nil (cons mov solucion)))
    ; Salir cuando encontremos solucion, devolviendola
    ((sol-val-soko est) (vis-soko est) (reverse solucion))
    (vis-soko est)
    ; Iterar hasta entrar un movimiento legal
    (loop 
     (format t "~%Movimiento (k=N,j=S,l=E,h=O)? ")
     (setf mov (car2num (read)))
     (if (mov-val mov est)
	 (return)
       (progn
	 (vis-soko est)
	 (format t "~%~s: movimiento no valido" mov))))))


; Para pruebas:
; -------------
;
#|
(load "sokoban")
(load "agente-humano")

(inic-soko #'agente-humano :verb t)

(describe 'inic-soko)
(describe 'gen-soko)
(describe 'soko)
(describe 'sol-val-soko)
(describe 'vis-soko)

(gen-soko)
(vis-soko (gen-soko))

(inic-soko #'agente-humano :verb t)
(inic-soko #'agente-humano :verb t :tipo 1)
(inic-soko #'agente-humano :verb t :tipo 2)
(inic-soko #'agente-humano :verb t :tipo 3)
(inic-soko #'agente-humano :verb t :tipo "screen.1")
(inic-soko #'agente-humano :verb t :tipo "kids02.scr")
(inic-soko #'agente-humano :verb t :tipo "kids09.scr")

|#
